/*
 * Framemaker MIF Output
 *
 * Ken Kundert
 */

#include <stdio.h>
#include <X11/Xlib.h>
#include "hard_devices.h"
#include "xgout.h"

#define VDPI			1200.0
#define MICRONS_PER_INCH	2.54E+04
#define POINTS_PER_INCH		72

typedef struct {
	char	*title_font;
	char	*axis_font;
	int	title_size;
	int	axis_size;
	FILE	*strm;
	} Info;


/*
 * Hardcopy Interface for Xgraph
 *
 * Major differences from first version:
 *   Four new parameters are passed to the device initialization routine:
 *   title_family, title_size, axis_family, and axis_size.  See the 
 *   description of xg_init() for details.
 *   
 *   Clipping is done automatically by xgraph.  The xg_clip() routine
 *   is obsolete.
 *
 *   The xg_line() routine has become the xg_seg() routine.  It now
 *   draws segments rather than a series of lines.  
 * 
 *   A new field (max_segs) in the device structure now specifies 
 *   the maximum number of segments the device can handle in a group.
 */


/*
 * Adding an output device to xgraph
 *
 * Step 1
 *   Write versions of the following routines for your device:
 *   xg_init(), xg_text(), xg_seg(), xg_dot(), and xg_end().
 *   The interface and function of these routines are described
 *   in detail below.  These routines should be named according
 *   to your device.  For example,  the initialization routine
 *   for the Postscript output device is psInit().  Also,  name
 *   your source file after your device (e.g. the postscript
 *   routines are in the file ps.c).  Instructions continue
 *   after the description of the interface routines.
 */


void fmText( );
void fmDot( );
void fmSeg( );
void fmEnd( );

int fmInit(strm, width, height, title_family, title_size,
            axis_family, axis_size, flags, out_info, errmsg)
	FILE *strm;			/* Output stream              */
	int width, height;		/* Size of space (microns)    */
	char *title_family;		/* Name of title font family  */
	double title_size;		/* Title font height (points) */
	char *axis_family;		/* Name of axis font family   */
	double axis_size;		/* Axis font height (points)  */
	int flags;		        /* Flags                      */
	xgOut *out_info;		/* Device info (RETURN)       */
	char errmsg[ERRBUFSIZE];	/* Error message area         */
/*
 * This routine is called by xgraph just before drawing is to
 * begin.  The desired size of the plot is given by `width'
 * and `height'.  The parameters `title_family', `title_size',
 * `axis_family', and `axis_size' specify the names of the
 * title and axis fonts and their vertical sizes (in points).
 * These parameters can be ignored if your device does not
 * support multiple fonts.  Binary flags are specified in
 * the `flags' field.  These include:
 *  D_DOCU:
 *      If this flag is set,  it indicates the user has specified that
 *	the output will be included in some larger document.  Devices
 *	may choose to use this information to produce output that
 *	can be integrated into documents with less effort.  For example,
 *	the Postscript output routines produce bounding box information
 *	when this flag is set.
 * The routine should fill in all of the fields of `out_info' with 
 * appropriate values.  The values are described below:
 *  area_w, area_h:  
 * 	Size of the drawing space in device coordinates.
 *	This should take in account the requested area
 *	given by `width', and `height'.
 *  bdr_pad:  
 * 	Xgraph will leave this number of device coordinates around
 *	all of the outer edges of the graph.
 *  axis_pad: 
 *	Additional space around axis labels (in devcoords)
 *	so that the labels do not appear crowded.
 *  legend_pad:
 *	Space (in devcoords) from the top of legend text to
 *	the representative line drawn above the legend text.
 *  tick_len:    
 *	Size of a tick mark placed on axis (in devcoords)
 *  axis_width:  
 *	An estimate of the width of a large character in
 *      the axis font (in devcoords).  This can be an overestimate.  An
 *      underestimate may produce bad results.
 *  axis_height: 
 *	An estimate of the height of a large character in
 *      the axis labeling font (in devcoords).
 *  title_width, title_height:  
 *	Same as above except for the title font.
 *  max_segs:
 *	Due to buffering constraints,  some devices may not be able to
 *	handle massive segment lists.  This parameter tells xgraph not
 *	to send more than `max_segs' segments in one request.
 * Output to the device should be written to the stream `strm'.
 * The functions are described individually below.  After filling
 * in the parameters and setting the function pointers,  the routine
 * should initialize its drawing state and store any extra needed
 * information in `user_state'.  This value will be passed to all
 * other routines during the drawing sequence.  If the device
 * cannot initialize,  it should return a zero status and fill
 * `errmsg' with an informative error message.
 */
{
	Info	*fm_info;
	char	**l;

	fm_info = (Info *) malloc(sizeof(*fm_info));

	fprintf(strm, "<MIFFile 3.0> # generated by Xgraph\n");
	fprintf(strm, "<Units Uin>\n");

	out_info->dev_flags = 0;
	out_info->bdr_pad = 0.0 * title_size * VDPI / POINTS_PER_INCH;
	out_info->axis_pad = 2.0 * axis_size * VDPI / POINTS_PER_INCH;
	out_info->legend_pad =  -0.2 * axis_size * VDPI / POINTS_PER_INCH;

	out_info->area_w = 1.5*(double)width * VDPI / MICRONS_PER_INCH;
	out_info->area_h = 1.5*(double)height  * VDPI/ MICRONS_PER_INCH;

	out_info->tick_len = 0.5 * axis_size * VDPI / POINTS_PER_INCH;
	out_info->axis_height = axis_size * VDPI / POINTS_PER_INCH;
	out_info->title_height = title_size * VDPI / POINTS_PER_INCH;
	out_info->axis_width = (5.0/12.0) * axis_size * VDPI / POINTS_PER_INCH;
	out_info->title_width =(5.0/12.0) * title_size * VDPI / POINTS_PER_INCH;
	out_info->max_segs = 1000;
	out_info->xg_text = fmText;
	out_info->xg_seg = fmSeg;
	out_info->xg_dot = fmDot;
	out_info->xg_end = fmEnd;
	out_info->user_state = (char *) fm_info;

	fm_info->title_font = title_family;
	fm_info->axis_font = axis_family;
	fm_info->title_size = title_size;
	fm_info->axis_size = axis_size;
	fm_info->strm = strm;
	return 1;
}


char *
fmEscapeString( to, from )

char *to, *from;
{
char *oto = to;

    while (*from != '\0')
    {   if (*from == '<' || *from == '>' || *from == '`' || *from == '\'') {
	    *to++ = '\\';
	}
	*to++ = *from++;
    }
    *to = '\0';
    return oto;
}




void fmText(user_state, x, y, text, just, style)
char *user_state;		/* Value set in xg_init   */
int x, y;			/* Text position (pixels) */
char *text;			/* Null terminated text   */
int just;			/* Justification (above)  */
int style;			/* Text style (above)     */
/*
 * This routine should draw text at the indicated position using
 * the indicated justification and style.  The justification refers
 * to the location of the point in reference to the text.  For example,
 * if just is T_LOWERLEFT,  (x,y) should be located at the lower left
 * edge of the text string.
 */
{
	char	*font;
	char	*halign, *valign;
	char	*bold;
	char	*angle;
	int	size;
	Info	*fm = (Info *) user_state;
	double	yoff;
	char	buf[BUFSIZ];

	(y);
	font = style == T_AXIS ? fm->axis_font:
		fm->title_font;
	size = style == T_AXIS ? fm->axis_size:
		fm->title_size;
	bold = style == T_AXIS ? "No": "Yes";
	angle = "Regular";

	fprintf(fm->strm, "<Font\n" );
	fprintf(fm->strm, "  <FFamily `%s'>\n", font);
	fprintf(fm->strm, "  <FSize  %d>\n", size);
	fprintf(fm->strm, "  <FBold  %s>\n", bold);
	fprintf(fm->strm, "  <FAngle  %s>\n", angle);
	fprintf(fm->strm, ">\n" );
	fprintf(fm->strm, "<TextLine\n", font, size);
	fprintf(fm->strm, "  <GroupID 1>\n");
	fprintf(fm->strm, "  <Angle 0>\n");
	switch (just) {
	    case T_CENTER:     halign = "Center"; yoff = size/3.0; break;
	    case T_LEFT:       halign = "Left";   yoff = size/3.0; break;
	    case T_RIGHT:      halign = "Right";  yoff = size/3.0; break;
	    case T_UPPERLEFT:  halign = "Left";   yoff = 0.0;      break;
	    case T_TOP:        halign = "Center"; yoff = 0.0;      break;
	    case T_UPPERRIGHT: halign = "Right";  yoff = 0.0;      break;
	    case T_BOTTOM:     halign = "Center"; yoff = 0.0;      break;
	    case T_LOWERLEFT:  halign = "Left";   yoff = size;     break;
	    case T_LOWERRIGHT: halign = "Right";  yoff = size;     break;
	    default:           halign = NULL;     yoff = 0.0;      break;
	}
	yoff /= POINTS_PER_INCH;
	if (halign)
	    fprintf(fm->strm, "  <TLAlignment %s>\n", halign);
	fprintf(fm->strm, "  <TLOrigin %f %f>\n", x/VDPI, y/VDPI + yoff);
	fmEscapeString( buf, text );
	fprintf(fm->strm, "  <String `%s'>\n", buf);
	fprintf(fm->strm, ">\n");
	return;
}


void fmSeg(user_state, ns, seglist, width, style, lappr, color)
char *user_state;		/* Value set in xg_init */
int ns;				/* Number of segments   */
XSegment *seglist;		/* X array of segments  */
int width;			/* Width of lines       */
int style;			/* See above            */
int lappr;			/* Line appearence      */
int color;			/* Line color (if any)  */
/*
 * This routine draws a number of line segments at the points
 * given in `seglist'.  Note that contiguous segments need not share
 * endpoints but often do.  All segments should be `width' devcoords wide
 * and drawn in style `style'.  If `style' is L_VAR,  the parameters
 * `color' and `lappr' should be used to draw the line.  Both
 * parameters vary from 0 to 7.  If the device is capable of
 * color,  `color' varies faster than `style'.  If the device 
 * has no color,  `style' will vary faster than `color' and
 * `color' can be safely ignored.  However,  if the
 * the device has more than 8 line appearences,  the two can
 * be combined to specify 64 line style variations.
 * Xgraph promises not to send more than the `max_segs' in the
 * xgOut structure passed back from xg_init().
 */
{
	Info	*fm = (Info *) user_state;
	char	*type, *color_name;
	int	i, j, k;
	double	linewidth;

	if (style == L_AXIS) {
	    type = "tick";
	    linewidth = 0.5;
	    color = 0;
	}
	else if (style == L_ZERO) {
	    type = "axis";
	    linewidth = 1.0;
	    color = 0;
	}
	else {
	    type = "trace";
	    linewidth = 2.0;
	    color = (lappr % 6) + 2;
	}

	for (i = 0; i < ns; i = j) {

		for (j = i + 1; j < ns
			&& seglist[j - 1].x2 == seglist[j].x1
			&& seglist[j - 1].y2 == seglist[j].y1;
			j++) ;

		fprintf(fm->strm, "<PolyLine # %s\n", type);
		fprintf(fm->strm, "  <GroupID 1>\n");
		fprintf(fm->strm, "  <Pen 0>\n");
		fprintf(fm->strm, "  <Fill 15>\n");
		fprintf(fm->strm, "  <Separation %d>\n", color);
		fprintf(fm->strm, "  <PenWidth %0.2f pt>\n", linewidth);
		fprintf(fm->strm, "  <HeadCap Round>\n");
		fprintf(fm->strm, "  <TailCap Round>\n");
		fprintf(fm->strm, "  <NumPoints %d>\n", j+1);
		for (k = i; k < j; k++) {
			fprintf(fm->strm, "  <Point %f\" %f\">\n",
				seglist[k].x1 / VDPI, seglist[k].y1 / VDPI);
		}
		fprintf(fm->strm, "  <Point %f\" %f\">\n",
			seglist[k-1].x2 / VDPI, seglist[k-1].y2 / VDPI);
		fprintf(fm->strm, ">\n");
	}
	return;
}

/* Marker styles */
#define P_PIXEL		0
#define P_DOT		1
#define P_MARK		2

void fmDot(user_state, x, y, style, type, color)
char *user_state;		/* Value set in xg_init    */
int x, y;			/* Location in pixel units */
int style;			/* Dot style               */
int type;			/* Type of marker          */
int color;			/* Marker color (if any)   */
/*
 * This routine should draw a marker at location `x,y'.  If the
 * style is P_PIXEL,  the dot should be a single pixel.  If
 * the style is P_DOT,  the dot should be a reasonably large
 * dot.  If the style is P_MARK,  it should be a distinguished
 * mark which is specified by `type' (0-7).  If the output
 * device is capable of color,  the marker should be drawn in
 * `color' (0-7) which corresponds with the color for xg_line.
 */
{
	Info	*fm = (Info *) user_state;
	char	*dotType;
	double	linewidth;
	double	X, Y, nudge;

	if (style == P_PIXEL) {
	    linewidth = 0.5;
	    dotType = "pixel";
	}
	else if (style == P_DOT) {
	    linewidth = 2.0;
	    dotType = "dot";
	}
	else if (style == P_MARK) {
	    linewidth = 4.0;
	    dotType = "mark";
	}

	X = ((double)x)/VDPI;
	Y = ((double)y)/VDPI;
	nudge = 0.25 / POINTS_PER_INCH;

	fprintf(fm->strm, "<PolyLine # %s\n", dotType);
	fprintf(fm->strm, "  <GroupID 1>\n");
	fprintf(fm->strm, "  <Pen 0>\n");
	fprintf(fm->strm, "  <Fill 15>\n");
	fprintf(fm->strm, "  <Separation %d>\n", color);
	fprintf(fm->strm, "  <PenWidth %0.2f pt>\n", linewidth);
	fprintf(fm->strm, "  <HeadCap Round>\n");
	fprintf(fm->strm, "  <TailCap Round>\n");
	fprintf(fm->strm, "  <NumPoints 2>\n");
	fprintf(fm->strm, "  <Point %f\" %f\">\n", X - nudge, Y - nudge);
	fprintf(fm->strm, "  <Point %f\" %f\">\n", X + nudge, Y + nudge);
	fprintf(fm->strm, ">\n");

	fprintf(fm->strm, "<PolyLine # %s\n", dotType);
	fprintf(fm->strm, "  <GroupID 1>\n");
	fprintf(fm->strm, "  <Pen 0>\n");
	fprintf(fm->strm, "  <Fill 15>\n");
	fprintf(fm->strm, "  <Separation %d>\n", color);
	fprintf(fm->strm, "  <PenWidth %0.2f pt>\n", linewidth);
	fprintf(fm->strm, "  <HeadCap Round>\n");
	fprintf(fm->strm, "  <TailCap Round>\n");
	fprintf(fm->strm, "  <NumPoints 2>\n");
	fprintf(fm->strm, "  <Point %f\" %f\">\n", X + nudge, Y - nudge);
	fprintf(fm->strm, "  <Point %f\" %f\">\n", X - nudge, Y + nudge);
	fprintf(fm->strm, ">\n");
	return;
}

void fmEnd(user_state)
char *user_state;
/*
 * This routine is called after a drawing sequence is complete.
 * It can be used to clean up the user state and set the device
 * state appropriately.  This routine is optional in the structure.
 */
{
	Info	*fm = (Info *) user_state;

	fprintf(fm->strm, "<Group <ID 1>>\n");
	fclose(fm->strm);
}

/*
 * Adding an output device to xgraph
 *
 * Step 2
 *   Edit the file hard_devices.c.  Declare your initialization
 *   function and add your device to the list of devices,
 *   hard_devices[].  The structure hard_dev is described below:
 */

#ifdef notdef
extern int fmInit( );

struct hard_dev fm = {
	"fm format", fmInit,
	0, ".clipboard", 0,
	25, "Times-Bold", 18, "Times", 12
	};
#endif

/*
 * dev_spec:
 *    The dev_spec field should be a command that directly outputs to
 *    your device.  The command should contain one %s directive that
 *    will be filled in with the name of the device from the hardcopy
 *    dialog.
 * dev_file:
 *    The default file to write output to if the user selects `To File'.
 * dev_printer:
 *    The default printer to write output to if the user selects
 *    `To Printer'.
 * dev_max_dim:
 *    The default maximum dimension for the device in centimeters.
 * dev_title_font, dev_title_size:
 *    The default title font and size.  Sizes are specified in
 *    points (1/72 inch).
 * dev_axis_font, dev_axis_size:
 *    The default axis font and size.
 */

/*
 * Adding an output device to xgraph
 *
 * Step 3
 *   Edit the file Makefile.  Add your source file to the SRC variable
 *   and the corresponding object file to the OBJ variable.  Finally,
 *   remake xgraph.  Your device should now be available in the
 *   hardcopy dialog.
 */

